#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/select.h>
#include <unistd.h>
#include <string.h>
#include <locale.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#define PROGRAM_NAME "vee"
#define PROGRAM_VERSION "0.1"
#define MIN_SIZE 4096

#ifdef DEBUG
#define debug(...) fprintf(stderr, __VA_ARGS__);
#else
#define debug(...)
#endif

void *buffer;
size_t buffer_size = MIN_SIZE;
void *read_c;
void *min_write_c;
int min_write_c_dirty;
struct writer { int fd; void *write_c; };
struct writer **writers;

int writer_add(struct writer ***writers, int fd) {
    if (!writers) return -1;

    debug("Adding writer for fd %d\n", fd);

    if (!(*writers)) {
        /* Crate first writer */
        *writers = malloc(sizeof(struct writer*[2]));
        if (!*writers)
            return -1;

        if (!( (*writers)[0] = malloc(sizeof(struct writer)) ))
            return -1;
        (*writers)[0]->fd = fd;
        (*writers)[0]->write_c = buffer;
        (*writers)[1] = (struct writer*)NULL;

    } else {
        /* Append next writer */
        int count = 0;
        struct writer **pwriter;
        
        /* Count old members without terminating NULL */
        for (pwriter=*writers; *pwriter; pwriter++) count++;
        
        /* Prolong the list */
        pwriter = realloc(*writers, sizeof(struct writer*[count + 2]));
        if (!pwriter) {
            perror("Appending writer into list via realloc() failed");
            return(-1);
        }
        *writers = pwriter;

        /* Write new writer down */
        if (!( (*writers)[count] = malloc(sizeof(struct writer)) ))
            return -1;
        (*writers)[count]->fd = fd;
        (*writers)[count]->write_c = buffer;
        (*writers)[count+1] = (struct writer*)NULL;

    }
    return 0;
}

void usage(void) {
    printf("Usage: %s [OPTION] [COMMAND [ARG...]]\n", PROGRAM_NAME);
    printf("Copy stdin to stdout and to stdin of COMMAND, if supplied.\n");
    printf("\n"
            "\t-s SIZE\t\tset size of buffer in bytes (default %zu B)\n"
            "\t-h\t\tprint this help message and exits\n"
            "\t-v\t\tprint version information and exits\n"
            , MIN_SIZE);
    printf("\n"
"Intermediate data are buffered. If buffer is full, no more data will be\n"
"read until same part of data will be written to all outputs, making\n"
"corresponding part of buffer free for new input data.\n"
            );
    printf("\n"
"If an output blocks, remaining ouputs will continue to be feed until\n"
"whole buffer will be chocked up with data waiting for blocked output.\n"
            );
}

void version(void) {
    printf("%s version %s\n"
        "Copyright (c) 2008 Petr Pisar\n"
        "License GPLv2+: GNU GPL version 2 or later "
            "<http://gnu.org/licenses/gpl.html>\n"
        "This is free software: you are free to change and redistribute it.\n"
        "There is NO WARRANTY, to the extent permitted by law.\n",
        PROGRAM_NAME, PROGRAM_VERSION);
}


/* Returns pointer to COMMAND in argv[] */
char **parse_args(int argc, char * const argv[]) {
    int opt;
    extern int optind;
    extern char *optarg;
    long int value;

    if (argc == 1) return NULL;

    while ((opt=getopt(argc, argv, "+hvs:")) != -1) {
        switch (opt) {
            case 'h':
                usage();
                exit(EXIT_SUCCESS);
                break;
            case 'v':
                version();
                exit(EXIT_SUCCESS);
                break;
            case 's':
                errno=0;
                value = strtoul(optarg, NULL, 0);
                if (errno || value == 0 || value <0) {
                    fprintf(stderr, "Option -s: invalid value: `%s'\n",
                            optarg);
                    exit(EXIT_FAILURE);
                }
                /* TODO: Exit if value can not fit into size_t */
                buffer_size=value;
                break;
            default:
                usage();
                exit(EXIT_FAILURE);
        }
    }

    return (optind<argc) ? (char **)&argv[optind] : NULL;
}

/* Waits for zombie and terminate main loop */
void terminateonchild(int signo) {
    int status;
    pid_t pid;

    pid = waitpid(-1, &status, WNOHANG);

    if (pid > 0 && (WIFEXITED(status) || WIFSIGNALED(status))) {
        fprintf(stderr, "Command exited ");
        if (WIFEXITED(status))
            fprintf(stderr, "with exit code %d\n", WEXITSTATUS(status));
        else
            fprintf(stderr, "with signal %d\n", WTERMSIG(status));
        fflush(stderr);
        exit(EXIT_SUCCESS);
    }

    if (SIG_ERR == signal(SIGCHLD, terminateonchild)) {
        perror("Can not reinstall SIGCHLD handler");
        exit(EXIT_FAILURE);
    }
}

/* Returns read site of pipe connected to forked and executed command from
 * argv[]. */
int pfopen(char *const argv[]) {
    int fd[2];
    pid_t child;
    int saved_errno;

    if (pipe(fd) == -1) {
        perror("pipe()");
        return -1;
    }

    /* install SIGCHLD handler to wait for zombie */
    if (SIG_ERR == signal(SIGCHLD, terminateonchild)) {
        perror("Can not install SIGCHLD handler");
        close(fd[0]);
        close(fd[1]);
        return -1;
    }

    child = fork();
    
    if (child == -1) {
        perror("fork()");
        close(fd[0]);
        close(fd[1]);
        return -1;
    }

    if (child == 0) {
        close(fd[1]);
        if (-1 == dup2(fd[0], STDIN_FILENO)) {
            perror("Can not redirect stdin");
            return -1;
        }

        execvp(argv[0], argv);

        saved_errno=errno;
        fprintf(stderr, "execvp(%s ...) failed: %s\n", argv[0],
                strerror(saved_errno));
        _exit(EXIT_FAILURE);
    } else {
        close(fd[0]);
    }

    return(fd[1]);
}

int main(int argc, char **argv, char **envp) {

    fd_set rfds;
    fd_set wfds;
    int max_fd;
    int retval=0;
    int saved_errno;
    int terminate=0;
    int input_closed=0;

    ssize_t processed;
    struct writer **pwriter;
    struct writer *awriter;

    char **command;
    int child_pipe;

    setlocale(LC_ALL, "");

    command = parse_args(argc, argv);
    
    /* Build buffer */
    if (!(buffer=malloc(buffer_size))) {
        fprintf(stderr, "Can not allocate buffer\n");
        exit(EXIT_FAILURE);
    }

    /* Register stdout */
    if (writer_add(&writers, STDOUT_FILENO)) {
        fprintf(stderr, "Internal error: Cannot build list of writers\n");
        exit(EXIT_FAILURE);
    }

    /* Execute command */
    if (command) {
        debug("Command supplied: %s\n", command[0]);
        child_pipe = pfopen(command);

        if (child_pipe == -1) {
            fprintf(stderr, "Command execution failed");
            exit(EXIT_FAILURE);
        }

        if (writer_add(&writers, child_pipe)) {
            fprintf(stderr,
                    "Internal error: Cannot add child to list of writers\n");
            exit(EXIT_FAILURE);
        }
    }


    /* Init cursors */
    read_c = buffer;
    min_write_c = buffer;
    
    /* Main loop */
    while (!terminate) {

        max_fd=0;
        /* Init rfds */
        FD_ZERO(&rfds);
        if (!input_closed && read_c < buffer+buffer_size) {
            FD_SET(0, &rfds);
            max_fd=1;
        } else {
            debug("Will not read\n");
        }
        
        /* Init wfds */
        FD_ZERO(&wfds);
        for (pwriter = writers; *pwriter; pwriter++) {
            awriter = *pwriter;
            if (awriter->write_c < read_c) {
                FD_SET(awriter->fd, &wfds);
                if (awriter->fd +1 > max_fd) max_fd = awriter->fd + 1;
            } else {
                debug("Will not write into fd %d\n", awriter->fd);
            }
        }

        /* Which FD is ready? */
        retval = select(max_fd, &rfds, &wfds, NULL, NULL);

        if (retval == -1) {
            if (errno != EINTR) {
                perror("select()");
                terminate=-1;
            }
            continue;
        }

        /* Reading */
        if (FD_ISSET(0, &rfds)) {
            processed = read(0, read_c, buffer_size - (read_c - buffer));
            if (processed < 0) {
                perror("Error while reading from stdin");
                terminate=-1;
            }
            else if (processed == 0) {
                fprintf(stderr, "EOF on stdin\n");
                input_closed=1;
            }
            else {
                debug("%d B read\n", processed);
                read_c+=processed;
            }
        }

        /* Writing */
        for (pwriter = writers; *pwriter; pwriter++) {
            awriter = *pwriter;
            if (FD_ISSET(awriter->fd, &wfds)) {
                processed = write(awriter->fd, awriter->write_c,
                        read_c - awriter->write_c);
                if (processed < 0) {
                    saved_errno = errno;
                    fprintf(stderr, "Error while writing into fd %d: %s",
                            awriter->fd, strerror(saved_errno));
                    terminate=-1;
                }
                else if (processed == 0) {
                    fprintf(stderr, "Nothing written into fd %d: "
                            "This should not happen\n", awriter->fd);
                }
                else {
                    debug("%d B written into fd %d\n", processed, awriter->fd);
                    awriter->write_c+=processed;
                    min_write_c_dirty=1;
                }
            }
        }

        
        /* Locate real garbage which has been written to all writers yet*/
        if (min_write_c_dirty) {
            min_write_c_dirty=0;

            min_write_c = buffer + buffer_size;
            for (pwriter = writers; *pwriter; pwriter++) {
                awriter = *pwriter;
                if (awriter->write_c < min_write_c)
                    min_write_c = awriter->write_c;
            }
        }

        /* Recycle buffer if there is no space for reading*/
        if (read_c == buffer+buffer_size) {
            /* Can ve move data */
            ssize_t unused;
            if (unused = min_write_c - buffer) {
                /* Yes, we can do it */
                memmove(buffer, min_write_c, read_c - min_write_c);
                min_write_c=buffer;
                read_c-=unused;
                for (pwriter = writers; *pwriter; pwriter++) {
                    (*pwriter)->write_c=buffer;
                }
                debug("Buffer recycled\n");
            } else
                debug("Buffer full\n");
        }

#ifdef DEBUG
        /* Buffer utilization */
        debug(  "buffer_size=%zu " \
                "to_read=%zd " \
                "to_write=%zd " \
                "unused=%zd " \
                "\n", \
                buffer_size, \
                buffer_size - (read_c-buffer), \
                read_c - min_write_c, \
                min_write_c - buffer \
                );
        for (pwriter = writers; *pwriter; pwriter++) {
            awriter = *pwriter;
            debug("fd=%d to_write=%zd\n", awriter->fd,
                    read_c - awriter->write_c);
        }
#endif


        /* Has all job been done? */
        if (input_closed && min_write_c == read_c) {
            fprintf(stderr, "All data written and no more input. Exiting.\n");
            terminate=1;
        }

    }


    free(buffer);
    exit((terminate<0) ? EXIT_FAILURE : EXIT_SUCCESS);
}
